// Copyright (C) 2006-2009 Simula Research Laboratory and Johan Hake
// Licensed under the GNU GPL Version 3.0.

#include <cstdlib>
#include <vector>

#include <tritetmesh/tetmesh/plceditor.h>

using namespace std;
using namespace tritetmesh;

int main(int argc, char *argv[])
{
  //TetgenBuilder tetbuilder;
  PLCEditor plceditor;
  
  // Add the coordinates for the vertices
  vector<double> xv, yv, zv;
  
  xv.push_back(0.0); xv.push_back(1.0); xv.push_back(0.0); xv.push_back(1.0);
  xv.push_back(0.0); xv.push_back(1.0); xv.push_back(0.0); xv.push_back(1.0);
  xv.push_back(0.25); xv.push_back(0.75); xv.push_back(0.25); xv.push_back(0.75);
  xv.push_back(0.25); xv.push_back(0.75); xv.push_back(0.25); xv.push_back(0.75);
  xv.push_back(1.0); xv.push_back(1.0); xv.push_back(1.0); xv.push_back(1.0);
  xv.push_back(0.875); xv.push_back(0.875); xv.push_back(0.875); xv.push_back(0.875);
  xv.push_back(0.0); xv.push_back(0.0); xv.push_back(0.0); xv.push_back(0.0);
  
  yv.push_back(0.0); yv.push_back(0.0); yv.push_back(1.0); yv.push_back(1.0);
  yv.push_back(0.0); yv.push_back(0.0); yv.push_back(1.0); yv.push_back(1.0);
  yv.push_back(0.25); yv.push_back(0.25); yv.push_back(0.75); yv.push_back(0.75);
  yv.push_back(0.25); yv.push_back(0.25); yv.push_back(0.75); yv.push_back(0.75);
  yv.push_back(0.25); yv.push_back(0.75); yv.push_back(0.25); yv.push_back(0.75);
  yv.push_back(0.25); yv.push_back(0.75); yv.push_back(0.25); yv.push_back(0.75);
  yv.push_back(0.25); yv.push_back(0.75); yv.push_back(0.25); yv.push_back(0.75);
  
  zv.push_back(0.0); zv.push_back(0.0); zv.push_back(0.0); zv.push_back(0.0);
  zv.push_back(1.0); zv.push_back(1.0); zv.push_back(1.0); zv.push_back(1.0);
  zv.push_back(0.25); zv.push_back(0.25); zv.push_back(0.25); zv.push_back(0.25);
  zv.push_back(0.75); zv.push_back(0.75); zv.push_back(0.75); zv.push_back(0.75);
  zv.push_back(0.25); zv.push_back(0.25); zv.push_back(0.75); zv.push_back(0.75);
  zv.push_back(0.25); zv.push_back(0.25); zv.push_back(0.75); zv.push_back(0.75);
  zv.push_back(0.25); zv.push_back(0.25); zv.push_back(0.75); zv.push_back(0.75);

  // Register the vertices
  plceditor.register_vertices(xv, yv, zv);
  
  // Add vertex markers
  vector<unsigned int> vmrk1, vmrk2;
  vmrk1.push_back(0); vmrk1.push_back(1); vmrk1.push_back(2); vmrk1.push_back(3); 
  plceditor.add_vertex_marker(0, vmrk1);
  
  vmrk2.push_back(4); vmrk2.push_back(5); vmrk2.push_back(6); vmrk2.push_back(7); 
  plceditor.add_vertex_marker(2, vmrk2);
  
  // Register the different facets
  vector<unsigned int> f1, f2, f3, f4, f5, f6, f7, f8, f9, f10, f11, f12, f13, f14, f15, f16,
              f17, f18, f19, f20, f21, f22, f23;
  vector<double> facet_hole1, facet_hole2;

  // Facet with marker
  f1.push_back(0); f1.push_back(1); f1.push_back(3); f1.push_back(2); 
  plceditor.add_facet(f1,1);

  // Facet with marker and area constraint
  f2.push_back(4); f2.push_back(5); f2.push_back(7); f2.push_back(6); 
  plceditor.add_facet(f2,2,0.001); 

  // Facet with marker and a hole
  f3.push_back(0); f3.push_back(2); f3.push_back(6); f3.push_back(4); 
  f23.push_back(24); f23.push_back(25); f23.push_back(27); f23.push_back(26); 
  facet_hole1.push_back(0.0); facet_hole1.push_back(0.5); facet_hole1.push_back(0.5);
  plceditor.add_facet(f3,3); 
  plceditor.add_polygon_to_last_facet(f23,facet_hole1); 

  // Facet with marker and area constraint
  plceditor.add_facet(f23,7,0.001);

  // Facet with marker and a hole
  f4.push_back(1); f4.push_back(3); f4.push_back(7); f4.push_back(5); 
  f17.push_back(16); f17.push_back(17); f17.push_back(19); f17.push_back(18); 
  facet_hole2.push_back(1.0); facet_hole2.push_back(0.5); facet_hole2.push_back(0.5);
  plceditor.add_facet(f4,4); plceditor.add_polygon_to_last_facet(f17,facet_hole2);

  // Internal facets
  f18.push_back(20); f18.push_back(21); f18.push_back(23); f18.push_back(22); plceditor.add_facet(f18);
  f19.push_back(16); f19.push_back(18); f19.push_back(22); f19.push_back(20); plceditor.add_facet(f19);
  f20.push_back(18); f20.push_back(19); f20.push_back(23); f20.push_back(22); plceditor.add_facet(f20);
  f21.push_back(17); f21.push_back(19); f21.push_back(23); f21.push_back(21); plceditor.add_facet(f21);
  f22.push_back(16); f22.push_back(17); f22.push_back(21); f22.push_back(20);  plceditor.add_facet(f22);

  // Facet with marker
  f5.push_back(0); f5.push_back(1); f5.push_back(5); f5.push_back(4); 
  plceditor.add_facet(f5,5);

  // Facet with marker
  f6.push_back(2); f6.push_back(3); f6.push_back(7); f6.push_back(6); 
  plceditor.add_facet(f6,6);

  // Internal facets
  f7.push_back(0+8); f7.push_back(1+8); f7.push_back(3+8); f7.push_back(2+8); plceditor.add_facet(f7);
  f8.push_back(4+8); f8.push_back(5+8); f8.push_back(7+8); f8.push_back(6+8); plceditor.add_facet(f8);
  f9.push_back(0+8); f9.push_back(2+8); f9.push_back(6+8); f9.push_back(4+8); plceditor.add_facet(f9);
  f10.push_back(1+8); f10.push_back(3+8); f10.push_back(7+8); f10.push_back(5+8); plceditor.add_facet(f10);
  f11.push_back(0+8); f11.push_back(1+8); f11.push_back(5+8); f11.push_back(4+8); plceditor.add_facet(f11);
  f12.push_back(2+8); f12.push_back(3+8); f12.push_back(7+8); f12.push_back(6+8); plceditor.add_facet(f12);
  f13.push_back(0); f13.push_back(2); f13.push_back(10); f13.push_back(8); plceditor.add_facet(f13);
  f14.push_back(0); f14.push_back(1); f14.push_back(9); f14.push_back(8); plceditor.add_facet(f14);
  f15.push_back(1); f15.push_back(3); f15.push_back(11); f15.push_back(9); plceditor.add_facet(f15);
  f16.push_back(2); f16.push_back(3); f16.push_back(11); f16.push_back(10); plceditor.add_facet(f16);

  // Add a hole and a region
  vector<double> hole, region;
  hole.push_back(0.5); hole.push_back(0.5); hole.push_back(0.5);
  plceditor.add_hole(hole);
  
  region.push_back(0.5); region.push_back(0.5); region.push_back(0.125);
  plceditor.add_region(region, 5.0, 0.0001);
  
  // Build the mesh
  plceditor.build("result",true);

  // Old functionality outcommented:
  // FIXME: Consider reimplement?
  

  //// Test the plc for intersecting facets
  //plc.test();
  //
  //// Build the tetmesh
  //plc.print_tetgen_parameters();
  //plc.build_mesh(tetmesh);
  //
  //tetmesh.check_consistency();
  //
  //// Save the result in different formats
  //plc.save_tetgen("plc");
  //tetmesh.save_tetgen("result");
  //tetmesh.save_medit("result");
  //tetmesh.save_gid("result");
  //tetmesh.save_off("result");
  //tetmesh.save_dolfin("result");
  //
  //// Save to dolfin mesh
  ////dolfconv.save("result",tetmesh);
  //
  //// Read the just saved dolfin mesh and refine it
  //TetMesh tetmesh_from_dolfin;
  //
  //dolfconv.load("result",tetmesh_from_dolfin);
  //
  //tetmesh_from_dolfin.refine_mesh(tetmesh,"rq1.1");
  //tetmesh.save_tetgen("refined");
  
  std::exit(0);
}
